import logging
import atexit

# Set a default logging configuration
# If a configuration was set before this was imported this will do nothing.
logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")

from . import (
    get_logger,
    unregister_default_log_handler,
    register_default_log_handler,
    get_min_log_level,
)

# Connect to python's logging module
logger = logging.getLogger("amulet.logging")


def on_msg(level: int, msg: str) -> None:
    logger.log(level, msg)


if get_min_log_level() <= 10:
    get_logger().dispatch(10, "Connecting python logger.")

logging_token = get_logger().connect(on_msg)


def python_shutdown() -> None:
    if get_min_log_level() <= 10:
        get_logger().dispatch(10, "Disconnecting python logger.")
    # On python shutdown, disconnect the logger so it can't get called after python has shut down.
    get_logger().disconnect(logging_token)
    # Reconnect the default logger to get logging after python has shut down.
    register_default_log_handler()
    if get_min_log_level() <= 10:
        get_logger().dispatch(10, "Disconnected python logger.")


atexit.register(python_shutdown)

# Disconnect default logging configuration.
# This handler is needed for any logging needed before this script is run.
unregister_default_log_handler()

if get_min_log_level() <= 10:
    get_logger().dispatch(10, "Connected python logger.")



from __future__ import annotations

import contextlib
import typing

import amulet.utils.task_manager.cancel_manager

__all__: list[str] = [
    "Deadlock",
    "Lock",
    "LockNotAcquired",
    "OrderedLock",
    "RLock",
    "SharedLock",
    "ThreadAccessMode",
    "ThreadShareMode",
]

class Deadlock(RuntimeError):
    """
    An exception raised in some deadlock cases.
    """

class Lock:
    """
    A wrapper for std::mutex.
    """

    def __enter__(self) -> None: ...
    def __exit__(
        self, exc_type: typing.Any, exc_val: typing.Any, exc_tb: typing.Any
    ) -> None: ...
    def __init__(self) -> None: ...
    def acquire(self, blocking: bool = True) -> bool: ...
    def release(self) -> None: ...

class LockNotAcquired(RuntimeError):
    """
    An exception raised if the lock was not acquired.
    """

class OrderedLock:
    """
    This is a custom mutex implementation that prioritises acquisition order and allows parallelism where possible.
    The acquirer can define the required permissions for this thread and permissions for other parallel threads.
    It also supports cancelling waiting through a CancelManager instance.
    """

    def __call__(
        self,
        blocking: bool = True,
        timeout: typing.SupportsFloat = -1.0,
        cancel_manager: amulet.utils.task_manager.cancel_manager.AbstractCancelManager = ...,
        thread_mode: tuple[ThreadAccessMode, ThreadShareMode] = ...,
    ) -> contextlib.AbstractContextManager[None, bool | None]:
        """
        A context manager to acquire and release the lock.
        Thread safe.

        >>> lock: OrderedLock
        >>> with lock():
        >>>     # code with lock acquired
        >>> # the lock will automatically be released here

        Entering the context manager acquires the lock.
        If the lock could not be acquired :class:`LockNotAcquired` is raised.
        Exiting the context manager releases the lock.

        :param blocking:
            If true (default) entering the context manager will block until the lock is acquired, the timeout is reached or the task is cancelled.
            If false entering the context manager will immediately fail if the lock could not be acquired.
        :param timeout:
            The maximum number of seconds to block for when entering the context manager.
            Has no effect if blocking is False. Default is forever.
        :param task_manager: A custom object through which acquiring can be cancelled.
            This effectively manually triggers timeout.
            This is useful for GUIs so that the user can cancel an operation that may otherwise block for a while.
        :param thread_mode: The permissions for the current and other parallel threads.
        :return: contextlib.AbstractContextManager[None]
        """

    def __init__(self) -> None: ...
    def acquire(
        self,
        blocking: bool = True,
        timeout: typing.SupportsFloat = -1.0,
        cancel_manager: amulet.utils.task_manager.cancel_manager.AbstractCancelManager = ...,
        thread_mode: tuple[ThreadAccessMode, ThreadShareMode] = ...,
    ) -> bool:
        """
        Acquire the lock.
        Thread safe.

        With improper use this can lead to a deadlock.
        Only use this if you know what you are doing. Consider using the context manager instead

        :param blocking:
            If true (default) this will block until the lock is acquired, the timeout is reached or the task is cancelled.
            If false and the lock cannot be acquired on the first try, this returns False.
        :param timeout: The maximum number of seconds to block for. Has no effect if blocking is False. Default is forever.
        :param task_manager: A custom object through which acquiring can be cancelled.
            This effectively manually triggers timeout.
            This is useful for GUIs so that the user can cancel an operation that may otherwise block for a while.
        :param thread_mode: The permissions for the current and other parallel threads.
        :return: True if the lock was acquired otherwise False.
        """

    def release(self) -> None:
        """
        Release the lock.
        Must be called by the thread that locked it.
        Thread safe.

        Only use this if you know what you are doing. Consider using the context manager instead
        """

class RLock:
    """
    A wrapper for std::recursive_mutex.
    """

    def __enter__(self) -> None: ...
    def __exit__(
        self, exc_type: typing.Any, exc_val: typing.Any, exc_tb: typing.Any
    ) -> None: ...
    def __init__(self) -> None: ...
    def acquire(self, blocking: bool = True) -> bool: ...
    def release(self) -> None: ...

class SharedLock:
    """
    A wrapper for std::shared_mutex.
    """

    def __init__(self) -> None: ...
    def acquire_shared(self, blocking: bool = True) -> bool: ...
    def acquire_unique(self, blocking: bool = True) -> bool: ...
    def release_shared(self) -> None: ...
    def release_unique(self) -> None: ...
    def shared(self) -> contextlib.AbstractContextManager[None, bool | None]: ...
    def unique(self) -> contextlib.AbstractContextManager[None, bool | None]: ...

class ThreadAccessMode:
    """
    Members:

      Read : This thread can only read.

      ReadWrite : This thread can read and write.
    """

    Read: typing.ClassVar[
        ThreadAccessMode
    ]  # value = amulet.utils.lock.ThreadAccessMode.Read
    ReadWrite: typing.ClassVar[
        ThreadAccessMode
    ]  # value = amulet.utils.lock.ThreadAccessMode.ReadWrite
    __members__: typing.ClassVar[
        dict[str, ThreadAccessMode]
    ]  # value = {'Read': amulet.utils.lock.ThreadAccessMode.Read, 'ReadWrite': amulet.utils.lock.ThreadAccessMode.ReadWrite}
    def __eq__(self, other: typing.Any) -> bool: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: typing.SupportsInt) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: typing.Any) -> bool: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class ThreadShareMode:
    """
    Members:

      Unique : Other threads can't run in parallel.

      SharedReadOnly : Other threads can only read in parallel.

      SharedReadWrite : Other threads can read and write in parallel.
    """

    SharedReadOnly: typing.ClassVar[
        ThreadShareMode
    ]  # value = amulet.utils.lock.ThreadShareMode.SharedReadOnly
    SharedReadWrite: typing.ClassVar[
        ThreadShareMode
    ]  # value = amulet.utils.lock.ThreadShareMode.SharedReadWrite
    Unique: typing.ClassVar[
        ThreadShareMode
    ]  # value = amulet.utils.lock.ThreadShareMode.Unique
    __members__: typing.ClassVar[
        dict[str, ThreadShareMode]
    ]  # value = {'Unique': amulet.utils.lock.ThreadShareMode.Unique, 'SharedReadOnly': amulet.utils.lock.ThreadShareMode.SharedReadOnly, 'SharedReadWrite': amulet.utils.lock.ThreadShareMode.SharedReadWrite}
    def __eq__(self, other: typing.Any) -> bool: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: typing.SupportsInt) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: typing.Any) -> bool: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...



from typing import TypeVar, Any

T = TypeVar("T")


def dynamic_cast(obj: Any, cls: type[T]) -> T:
    """Like typing.cast but with runtime type checking."""
    if isinstance(obj, cls):
        return obj
    raise TypeError(f"{obj} is not an instance of {cls}")



from __future__ import annotations

import collections.abc
import typing

import amulet.utils.event

__all__: list[str] = [
    "AbstractProgressManager",
    "ProgressManager",
    "VoidProgressManager",
]

class AbstractProgressManager:
    def get_child(
        self, progress_min: typing.SupportsFloat, progress_max: typing.SupportsFloat
    ) -> AbstractProgressManager:
        """
        Get a child ProgressManager.
        If calling multiple functions, this allows segmenting the reported time.
        Thread safe.
        """

    def register_progress_callback(
        self, callback: collections.abc.Callable[[float], None]
    ) -> amulet.utils.event.EventToken[float]:
        """
        Register a function to get called when progress changes.
        The callback will be called from the thread `update_progress` is called in.
        Thread safe.
        """

    def register_progress_text_callback(
        self, callback: collections.abc.Callable[[str], None]
    ) -> amulet.utils.event.EventToken[str]:
        """
        Register a function to get called when progress changes.
        The callback will be called from the thread `update_progress` is called in.
        Thread safe.
        """

    def unregister_progress_callback(
        self, token: amulet.utils.event.EventToken[typing.SupportsFloat]
    ) -> None:
        """
        Unregister a registered function from being called when update_progress is called.
        Thread safe.
        """

    def unregister_progress_text_callback(
        self, token: amulet.utils.event.EventToken[str]
    ) -> None:
        """
        Unregister a registered function from being called when update_progress is called.
        Thread safe.
        """

    def update_progress(self, progress: typing.SupportsFloat) -> None:
        """
        Notify the caller of the updated progress.
        progress must be in the range 0.0 - 1.0
        Thread safe.
        """

    def update_progress_text(self, text: str) -> None:
        """
        Send a new progress text to the caller.
        Thread safe.
        """

class ProgressManager(AbstractProgressManager):
    def __init__(self) -> None: ...
    def __repr__(self) -> str: ...

class VoidProgressManager(AbstractProgressManager):
    def __init__(self) -> None: ...
    def __repr__(self) -> str: ...



from __future__ import annotations

import os
import pathlib

__all__: list[str] = ["get_temp_dir", "set_temp_dir"]

def get_temp_dir() -> pathlib.Path:
    """
    Get the directory in which temporary directories will be created.
    This is configurable by setting the "CACHE_DIR" environment variable.
    Thread safe.
    """

def set_temp_dir(path: os.PathLike | str | bytes) -> None:
    """
    Set the temporary directory path.
    It must be a path to an existing directory.
    Anything using the previous path will continue using that path.
    """
